# -*- coding: japanese-shift-jis-dos; tab-width: 4; -*-
#
# wdic-body.pl
# $Id: wdic-body.pl 1.5.0.1 2004/12/12 13:40:48 satomii Exp $
#
# This file is designed to be used from the wdic-fpw.pl script.
# DO NOT CHANGE THE KANJI CODE OF THIS FILE.
#
# Copyright (C) 2000-2004, rei@wdic.org.
# This program is distributed in the hope that it will be useful, but
# without any warranty. See the GNU General Public License for the details.
#
require 5.005;

use strict;
use IO::File;
use File::DosGlob 'glob';
use Jcode;

require 'wdic-reform.pl';

my %LABEL_EUC = (
	dir => Jcode::convert(': ', 'euc', 'sjis'),
	read => Jcode::convert('ǂ: ', 'euc', 'sjis'),
	spell => Jcode::convert('O: ', 'euc', 'sjis'),
	pron => Jcode::convert(': ', 'euc', 'sjis'),
	pos => Jcode::convert('i: ', 'euc', 'sjis'),
	plugin => Jcode::convert('yvOCz', 'euc', 'sjis'),
	graph => Jcode::convert('[}]', 'euc', 'sjis'),
	sound => Jcode::convert('[]', 'euc', 'sjis'),
);

#
# void make_body(int wc)
# the entry point to generate word entries.
#
sub make_body($)
{
	my $count = $_[0];
	my ($i, $key);

	&log("Writing $count word entries...\n");

	for ($i = 0; $i < $count; $i++) {
		my %word = $wdic->define_word($i);
		&add_word(%word);
	}

	$i += &write_plugin_images();
	$i += &write_plugin_sounds();
	$i += &write_plugin_files();
	$i += &write_plugin_ext();

	&log("Total $count words, $i entries written.\n");
}

#
# (int, int) new_entry(char *tag, int offset)
# switch to the new entry and writes a tag. it is recommended to add
# FreePWING::BaseText::add_entry_tag method to your FreePWING package
# by applying the patch from [edict 1364].
#
sub new_entry($;$)
{
	my ($tag, $offset) = @_;
	my ($text_pos, $head_pos) = (0, 0);

	$fpwtext->new_entry()
		|| die &fpwerr("Unable to start new entry", $fpwtext);
	$fpwheading->new_entry()
		|| die &fpwerr("Unable to start new entry", $fpwheading);
	$text_pos = $fpwtext->entry_position();
	$head_pos = $fpwheading->entry_position();

	$tag .= sprintf("%08x", $offset) if (defined($offset));

	if ($fpwtext->can('add_entry_tag')) {
		$fpwtext->add_entry_tag($tag)
			|| die &fpwerr("Unable to add entry tag: $tag", $fpwtext);
	} else {
		$fpwtext->add_tag($tag)
			|| die &fpwerr("Unable to add entry tag: '$tag", $fpwtext);
	}
	return ($head_pos, $text_pos, $tag);
}

##------------------------------------------------------------------------
## WORD ENTRY

#
# void add_word(%word)
# adds one word entry to the FreePWING context.
#
sub add_word(%)
{
	my %word = @_;
	my ($entry_tag, $text_pos, $head_pos);
	my (@list, $word, $group_name, $file_name, $index);
	local $_;

	printf("$word{GROUP}:$word{FILE}(%08x): %s\n",
		$word{OFFSET}, Jcode::convert($word{WORD}, $Wdic::CHARSET, 'sjis'));

	$word = $word{WORD};
	if ($word =~ s/(\s+\([0-9]+\))$//) { $index = $1; }
	else { $index = ''; }

	$group_name = $wdic->get_group_title($word{GROUP});
	$group_name =~ s/̊bm$//;
	$file_name = $wdic->get_file_title($word{FILE});
	$file_name = ":$file_name" if ($group_name);

	#
	# switch to the new entry and write the heading.
	#
	($head_pos, $text_pos, $entry_tag) = &new_entry($word{FILE}, $word{OFFSET});
	$_ = $group_name;
	$_ =~ s/p$//;
	$_ .= $file_name;
	$_ =~ s/$//;
	&write_sjis($fpwheading, "$word{WORD} y$_z");

	#
	# register search words.
	#
	$_ = &reform_sjis($word, 2);
	if (!/^[\s\-\']*$/) {
		$_ = Jcode::convert($_, 'euc', 'sjis');
		$fpwword2->add_entry($_, $head_pos, $text_pos)
			|| die &fpwerr("Unable to add search word: $_", $fpwword2);
	}
	push(@list, $word, $word{FILE});

	if ($word{READ} !~ /^\?+$/) {
		$fpwword2->add_entry(Jcode::convert($word{READ}, 'euc', 'sjis'),
				$head_pos, $text_pos)
			|| die &fpwerr("Unable to add search word: $word{READ}",
				$fpwword2);
		push(@list, &make_reading_keywords($word{READ}));
	}

	push(@list, $word{GROUP}) if ($word{GROUP} =~ /^[A-Z]+$/i);
	push(@list, &make_english_keywords($head_pos, $text_pos, $word{SPELL}));
	push(@list, &make_link_keywords(@{ $word{REFERENCE} }));
	push(@list, @{ $word{DIRS} }) if ($word{DIRS});
	&add_keyword_list($head_pos, $text_pos, @list);
	undef @list;

	#
	# write the first line.
	#
	if (!$fpwtext->add_keyword_start()
			|| !$fpwtext->add_text(Jcode::convert($word, 'euc', 'sjis'))
			|| !$fpwtext->add_keyword_end()) {
		die &fpwerr("Unable to write the first line", $fpwtext);
	}
	&write_sjis($fpwtext, "$index y$group_name$file_namez\n");

	#
	# write the directory (category) lines.
	#
	if ($opt_dirinfo) {
		&write_dir_link($fpwtext, 0, @{ $word{DIRS} });
	}

	#
	# write the next line which may contain word reading, english,
	# pronounce and POS.
	#
	$fpwtext->add_text($LABEL_EUC{read})
		|| die &fpwerr("Unable to write text: $LABEL_EUC{read}", $fpwtext);
	&write_sjis($fpwtext, "$word{READ}\n");

	&write_spell($word{SPELL}, $word{FILE}) if ('' ne $word{SPELL});
	&write_phonetic($word{PRONOUNCE}, $word{FILE}) if ('' ne $word{PRONOUNCE});

	if ('' ne $word{POS}) {
		$fpwtext->add_text($LABEL_EUC{pos})
			|| die &fpwerr("Unable to write text: $LABEL_EUC{pos}", $fpwtext);
		&write_sjis($fpwtext, "$word{POS}\n");
	}

	&write_newline($fpwtext, 1);
	$fpwtext->add_indent_level(2);

	# write definition lines.
	@list = @{ $word{DEFINITION} };
	if (@list) {
		foreach (@list) {
			warn "$word{FILE}: Unexpected format: $_\n" if (!s/^E//);
			$fpwtext->add_text($opt_bullet)
				|| die &fpwerr("Unable to add text: $opt_bullet", $fpwtext);
			$_ = &reform_sjis($_, 0);
			&write_sjis($fpwtext, "$_\n");
		}
	}

	#
	# write columns and/or examples, if any.
	#
	@list = @{ $word{COLUMN} };
	if (@list) {
		my $image_block = 0;
		&write_newline($fpwtext, 1) if (@{ $word{DEFINITION} });
		foreach (@list) {
			if (/^/) {
				$image_block = 1;
			} elsif (/^$/) {
				$image_block = 0;
			} elsif ($image_block && /^/) {
				&write_link_list($word{WORD}, $word{GROUP}, ( $_ ));
				next;
			}
			&write_sjis($fpwtext, "$_\n");
		}
	}

	#
	# write links.
	#
	@list = @{ $word{REFERENCE} };
	if (@list) {
		if (@{$word{DEFINITION} } || @{ $word{COLUMN} }) {
			&write_newline($fpwtext, 1);
		}
		&write_link_list($word{WORD}, $word{GROUP}, @list);
	}

	# and add a newline at the end of this entry.
	&write_newline($fpwtext, 1);
}

##------------------------------------------------------------------------
## KEYWORDS

#
# void add_keyword_list(char **euc_keywords)
# registers each keywords in the given list to the current word entry.
#
sub add_keyword_list($$@)
{
	my ($head_pos, $text_pos, @key) = @_;
	my ($key, $prev_key);

	foreach $key (@key) {
		# do not use DBCS alphabets - due to FreePWING behavior (bug?).
		$key = &reform_sjis($key, 2);
		$key = Jcode::convert($key, 'euc', 'sjis');
		$key = uc($key);
	}

	@key = sort(@key);
	$prev_key = '';

	foreach $key (@key) {
		next if ($key =~ /^[\s\-\']*$/ || $key eq $prev_key);
		$fpwkeyword->add_entry($key, $head_pos, $text_pos)
			|| die &fpwerr("Unable to add keyword: $key", $fpwkeyword);
		$prev_key = $key;
	}
}

#
# char **make_reading_keywords(char *read)
#
sub make_reading_keywords($)
{
	my $read = $_[0];
	my (@key, $key, $c);

	$key = '';
	while ('' ne $read) {
		$c = substr($read, 0, 1);
		$read = substr($read, 1);
		if ($c =~ /[\x81-\x9f\xe0-\xfc]/) {
			$c .= substr($read, 0, 1);
			$read = substr($read, 1);
		}
		if ('@' eq $c || 1 == length($c)) {	# '@' or '-'.
			push(@key, $key) if ('' ne $key);
			$key = '';
		} else {
			$key .= $c;
		}
	}
	push(@key, $key) if ('' ne $key);
	return @key;
}

#
# char **make_english_keywords(int heading_pos, int text_pos, char *english)
# registers english terms as search words. also returns the keyword list
# for 'conditional' search.
#
sub make_english_keywords($$$)
{
	my ($head_pos, $text_pos, $english) = @_;
	my (%keyword, @key);

	$english = &reform_sjis($english, 2);
	$english = Jcode::convert($english, 'euc', 'sjis');
	s/ *\[[^\]]+\]//g;
	s/\\/\xa1\xef/g;		# '' in EUC (a1ef).

	# search words.
	@key = split(/[:;\/] +/, $english);
	foreach (@key) {
		s/^\s+//;
		s/\s+$//;
		if (!/^[\-\']*$/ && !$keyword{$_}) {
			$fpwword2->add_entry($_, $head_pos, $text_pos)
				|| die &fpwerr("Cannot add search word: $_ (in \"$english\")",
						$fpwword2);
			$_ = uc($_);
			$keyword{$_} = 1;
		}
	}

	# jouken-kensaku keywords.
	$_ = Jcode::convert(join(' ', @key), 'sjis', 'euc');
	split(/\s+/);
}

#
# char **make_link_keywords(char **link_list)
#
sub make_link_keywords(@)
{
	my (@key, $key, $str, $c);

	foreach (@_) {
		next if (!/^\x81[\xa8\x9f\x9a]/);	# ![], not a 'word'.
		$str = $_;
		$str =~ s/^\x81.//;
		$str =~ s/ +:[A-Z]+$//i;
		$str =~ s/ +\([0-9]+\)$//;
		push(@key, $str);

		$key = '';
		while ('' ne $str) {
			$c = substr($str, 0, 1);
			$str = substr($str, 1);
			if ($c =~ /[\x81-\x9f\xe0-\xfc]/) {
				$c .= substr($str, 0, 1);
				$str = substr($str, 1);
			}
			if ('@' eq $c) {
				push(@key, $key) if ('' ne $key);
				$key = '';
			} else {
				$key .= $c;
			}
		}
		push(@key, $key) if ('' ne $key && $_ ne $key);
	}

	@key;
}

##------------------------------------------------------------------------
## REFERENCES

#
# void write_link_list(char *current_word, char *current_group, char **list)
# writes a list of related words, making cross-reference available if
# the target word is found in the loaded words.
#
sub write_link_list($$@)
{
	my $current_word = shift;
	my $current_group = shift;
	local $_;

	foreach (@_) {
		if (s/^([^ \/\\<>\|\*\?]+) *//i) {
			# plugin file.
			my $key = $1;
			my $plugin = $1;
			my $ext = '';
			my $group = '';

			if ($key =~ s/^(.+)\.([^\.]+)$/$2_$1/) {
				$ext = lc(".$2");
			}
			if (s/^:([A-Z]+) *//) {
				my $gid = $1;
				$group = $wdic->get_group_title($gid);
				$group = $gid if (!$group);
				$group =~ s/̊bm$//;
				$group =~ s/p$//;
			}

			$_ = &reform_sjis($_, 0);
			$_ = " $_" if ('' ne $_);
			$_ = " s$groupt$_" if ('' ne $group);

			if (defined($cgraphs{$key})) {
				# image plugin.
				push(@{ $cgraphs{$key}{list} }, "$current_word");
				$_ = "$plugin$_";
				if (!$fpwtext->add_color_graphic_start($key)
						|| !&write_sjis($fpwtext, $_)
						|| !$fpwtext->add_color_graphic_end()) {
					die &fpwerr("Unable to add DIB: $_", $fpwtext);
				}
				next;

			} elsif (defined($sounds{$key})) {
				# sound plugin.
				push(@{ $sounds{$key}{list} }, "$current_word");
				&write_sjis($fpwtext, "");
				if (!$fpwtext->add_sound_start($key)
						|| !&write_sjis($fpwtext, $plugin)
						|| !$fpwtext->add_sound_end()) {
					die &fpwerr("Unable to add sound: $key", $fpwtext);
				}
				&write_sjis($fpwtext, $_) if ('' ne $_);
				next;

			} elsif ('' eq $ext || $ext =~ /^\.(asm|csv|doc|pl|txt)$/) {
				# text plugin.
				my $file = &find_plugin($plugin);
				if ($file) {
					if (defined($plugins{$plugin})) {
						push(@{ $plugins{$plugin}{list} }, "$current_word");
					} else {
						$plugins{$plugin}{tag} = sprintf("PLUGIN%04x",
								scalar(keys(%plugins)) + 1);
						$plugins{$plugin}{path} = $file;
						$plugins{$plugin}{list} = [ "$current_word" ];
					}
					&write_sjis($fpwtext, '');
					&write_link($fpwtext, $plugin, $plugins{$plugin}{tag});
					&write_sjis($fpwtext, $_) if ('' ne $_);
					next;
				}
			}

			# target not available, default to text.
			&write_sjis($fpwtext, "$plugin$_");

		} elsif (s/^(\x81[\xa8\x9f\x9a])//) {
			# [], link to another entry.
			my $bullet = $1;
			my $gid = (s/ :([A-Z]+)$//i) ? $1 : $current_group;
			my ($group, $file) =
					&write_word_link($fpwtext, $bullet, $_, $gid);
			$group = $wdic->get_group_title($group ? $group : $gid);
			$group =~ s/̊bm$//;
			$group =~ s/p$//;
			if ($file) {
				$file = $wdic->get_file_title($file);
				$file =~ s/$//;
				$group .= ':' if (group);
				$group .= $file;
			}
			&write_sjis($fpwtext, " s$groupt") if ($group);

		} else {
			# not [], external information.
			if (s/^([^ ]+)(.*)$/$1/) {	# URL.
				if ($2) {
					$_ .= ' ' . &reform_sjis($2, 0);
				}
			} elsif (/^/) {				# URL group heading.
				$_ = &reform_sjis($_, 0);
			}
			&write_sjis($fpwtext, $_);
		}

	} continue {
		&write_newline($fpwtext, 1);
	}
}

##------------------------------------------------------------------------
## SYMBOLS

sub write_spell($$)
{
	local $_ = $_[0];
	my $file = $_[1];

	$fpwtext->add_text($LABEL_EUC{spell})
		|| die &fpwerr("$file: Unable to write text: $LABEL_EUC{spell}",
				$fpwtext);

	while ('' ne $_) {
		my $c = '';
		my @gaiji;

		if (s/^f//) {
			$c = '\'';
		} elsif (s/^($SJIS_ANY)//) {
			$c = Jcode::convert($1, 'euc', 'sjis');
		} elsif (s/^\]\[//) {
			$c = '][';
		} elsif (s/^ *\[//) {
			$c = ' [';
		} elsif (s/^ *; *//) {
			$c = '; ';
		} elsif (s/^ +//) {
			$c = ' ';
		} elsif (!$opt_gaiji) {
			# nop.
		} elsif (s/^\'//) {
			if (s/^([aeiou])//) {
				$gaiji[0] = "${1}u";
			} elsif (s/^([EI])//) {
				# 1.3c: currently `[AOU] does not appear anywhere.
				$gaiji[0] = 'c_' . lc($1) . 'u';
			} elsif (s/^'([aeou])//) {
				# ''[aeou], with umlaut.
				$gaiji[0] = "${1}uml";
			} else {
				$c = '\'';
				warn "$file: Unknown symbol: \'$_ (in $_[0])\n";
			}
		} elsif (s/^`//) {
			# 1.3c: currently `[AEIOU] does not appear anywhere.
			if (s/^([aeiou])//) {
				$gaiji[0] = "${1}d";
			} else {
				$c = '`';
				warn "$file: Unknown symbol: `$_ (in $_[0])\n";
			}
		} elsif (s/^\^//) {
			if (s/^([aceghjosu])//) {
				$gaiji[0] = "${1}h";
			} elsif (s/^([CGHJOSU])//) {
				$gaiji[0] = 'c_' . lc($1) . 'h';
			} elsif (!/^[0-9]/) {
				$c = '^';
				warn "$file: Unknown symbol: ^$_ (in $_[0])\n";
			}
		} elsif (s/^~//) {
			if (s/^u//) {
				$gaiji[0] = 'ufx';
			} elsif (s/^U//) {
				$gaiji[0] = 'c_ufx';
			} elsif (s/^n//) {
				# 1.3c: currently ~N does not appear anywhere.
				$gaiji[0] = 'nt';
			} else {
				$c = '~';
				warn "$file: Unknown symbol: ~$_ (in $_[0])\n";
			}
		} elsif (s/^\+S//) {
			# 1.3c: currently +s does not appear anywhere.
			$gaiji[0] = 'szlig';
		}

		if (@gaiji) {
			&write_gaiji($fpwtext, @gaiji);
		} else {
			if ('' eq $c) {
				$c = substr($_, 0, 1);
				$_ = substr($_, 1);
			}
			$fpwtext->add_text($c)
				|| die &fpwerr("$file: Unable to write text: $c (in $_[0])",
						$fpwtext);
		}
	}

	&write_newline($fpwtext, 1);
}

#
# void write_phonetic(char *pron, char *file);
# writes a pronounce into the current context using or not using user
# characters.
#
sub write_phonetic($$)
{
	local $_ = $_[0];
	my $file = $_[1];

	$fpwtext->add_text($LABEL_EUC{pron})
		|| die &fpwerr("$file: Unable to write text: $LABEL_EUC{pron}",
				$fpwtext);

	while ('' ne $_) {
		my $c = '';
		my @gaiji;

		if (s/^($SJIS_ANY)//) {
			$c = Jcode::convert($1, 'euc', 'sjis');
		} elsif (s/^\]\[//) {
			$c = '][';
		} elsif (s/^ *\[//) {
			$c = ' [';
		} elsif (s/^ *\| *//) {
			$c = ' | ';
		} elsif (s/^ +//) {
			$c = ' ';
		} elsif (!$opt_gaiji) {
			# nop.
		} elsif (s/^([3S])//) {
			$gaiji[0] = $1;
		} elsif (s/^0\-//) {
			$gaiji[0] = '0';
		} elsif (s/^(a)e// || s/^\{([ao])e\}//) {
			@gaiji = ("${1}e", 'ee');
		} elsif (s/^n,//) {
			$gaiji[0] = 'n';
		} elsif (s/^\'//) {
			if (s/^(a)e// || s/^\{([ao])e\}//) {
				@gaiji = ("${1}eu", 'eeu');
			} elsif (s/^([aeioSu])//) {
				$gaiji[0] = "${1}u";
			} elsif (s/^\#//) {
				if (s/^([cev])//) {
					$gaiji[0] = "r${1}u";
				} else {
					warn "$file: Unknown phonetic symbol: \'#$_ (in $_[0])\n";
				}
			} elsif (s/^6//) {
				$gaiji[0] = '6';
			} else {
				# &#768;
			}
		} elsif (s/^`//) {
			if (s/^(a)e// || s/^\{([ao])e\}//) {
				@gaiji = ("${1}ed", 'eed');
			} elsif (s/^([aeioSu])//) {
				$gaiji[0] = "${1}d";
			} elsif (s/^\#//) {
				if (s/^([cev])//) {
					$gaiji[0] = "r${1}d";
				} else {
					warn "$file: Unknown phonetic symbol: `#$_ (in $_[0])\n";
				}
			} else {
				# &#769;
			}
		} elsif (s/^\#//) {
			if (s/^e\+//) {
				$gaiji[0] = 'rei';
			} elsif (s/^([3acev])//) {
				$gaiji[0] = "r${1}";
			} else {
				warn "$file: Unknown phonetic symbol: #$_ (in $_[0])\n";
			}
		} elsif (s/^~//) {
			# french phonetic symbols.
			if (s/^([an])//) {
				$gaiji[0] = "${1}t";
			} elsif (s/^\#([3c])//) {
				$gaiji[0] = "r${1}t";
			} elsif (s/^\{oe\}//) {
				@gaiji = ('oet', 'eet');
			} else {
				warn "$file: Unknown phonetic symbol: ~$_ (in $_[0])\n";
			}
		} elsif (s/^\(r\)//) {
			$gaiji[0] = 'r';
		} elsif (/^\{[a-z]+\}/i) {
			warn "$file: Unknown phonetic symbol: $_ (in $_[0])\n";
		}

		if (@gaiji) {
			&write_gaiji($fpwtext, @gaiji);
		} else {
			if ('' eq $c) {
				$c = substr($_, 0, 1);
				$_ = substr($_, 1);
			}
			$fpwtext->add_text($c)
				|| die &fpwerr("$file: Unable to write text: $c (in $_[0])",
						$fpwtext);
		}
	}

	&write_newline($fpwtext, 1);
}

##------------------------------------------------------------------------
## PLUGIN FILE ENTRY

sub find_plugin($)
{
	my $plugin = "$opt_plugin/$_[0]";
	my $file = File::DosGlob::glob($plugin) || CORE::glob($plugin);
	if (!$file && -f $plugin) {
		$plugin;
	} else {
		$file;
	}
}

#
# int write_plugin_images(void)
# creates text entries for plugin images. image files that are not
# referred by any word entries are also listed here.
#
sub write_plugin_images()
{
	my @keys = keys(%cgraphs) or return 0;
	my ($key, $name, $head_pos, $text_pos);

	&log("Writing graphics list...\n");
	@keys = sort { $cgraphs{$a}{name} cmp $cgraphs{$b}{name}; } @keys;
	$fpwtext->new_context()
		|| die &fpwerr("Unable to start new context", $fpwtext);

	foreach $key (@keys) {
		$name = $cgraphs{$key}{name};
		$cgraphs{$key}{tag} = "GRAPH_$key";
		($head_pos, $text_pos) = &new_entry($cgraphs{$key}{tag});

		$fpwheading->add_text("$name $LABEL_EUC{plugin}")
			|| die &fpwerr("Unable to add heading: $name $LABEL_EUC{plugin}",
					$fpwheading);
		$fpwkeyword->add_entry($name, $head_pos, $text_pos)
			|| die &fpwerr("Unable to add keyword: $name", $fpwkeyword);

		if (!$fpwtext->add_keyword_start()
				|| !$fpwtext->add_text($name)
				|| !$fpwtext->add_keyword_end()
				|| !$fpwtext->add_newline()) {
			 die &fpwerr("Unable to write the first line", $fpwtext);
		}
		if (!$fpwtext->add_color_graphic_start($key)
				|| !$fpwtext->add_text($LABEL_EUC{graph})
				|| !$fpwtext->add_color_graphic_end()
				|| !$fpwtext->add_newline()) {
			die &fpwerr("Unable to add graphic: $key ($name)", $fpwtext);
		}
		if ($cgraphs{$key}{list}) {
			&write_link_list('', '', @{ $cgraphs{$key}{list} });
		}
		&write_newline($fpwtext, 1);
	}

	scalar(@keys);
}

#
# int write_plugin_sounds(void)
# creates text entries for plugin soudns. sound files that are not
# referred by any word entries are also listed here.
#
sub write_plugin_sounds()
{
	my @keys = keys(%sounds) or return 0;
	my ($key, $name, $head_pos, $text_pos);

	&log("Writing sounds list...\n");
	@keys = sort { $sounds{$a}{name} cmp $sounds{$b}{name} } @keys;
	$fpwtext->new_context()
		|| die &fpwerr("Unable to start new context", $fpwtext);

	foreach $key (@keys) {
		$name = $sounds{$key}{name};
		$sounds{$key}{tag} = "SOUND_$key";
		($head_pos, $text_pos) = &new_entry($sounds{$key}{tag});

		$fpwheading->add_text("$name $LABEL_EUC{plugin}")
			|| die &fpwerr("Unable to add heading: $name $LABEL_EUC{plugin}",
					$fpwheading);
		$fpwkeyword->add_entry($name, $head_pos, $text_pos)
			|| die &fpwerr("Unable to add keyword: $name", $fpwkeyword);

		if (!$fpwtext->add_keyword_start()
				|| !$fpwtext->add_text($name)
				|| !$fpwtext->add_keyword_end()
				|| !$fpwtext->add_newline()) {
			 die &fpwerr("Unable to write the first line", $fpwtext);
		}
		if (!$fpwtext->add_sound_start($key)
				|| !$fpwtext->add_text($LABEL_EUC{sound})
				|| !$fpwtext->add_sound_end()
				|| !$fpwtext->add_newline()) {
			die &fpwerr("Unable to add sound: $key ($name)", $fpwtext);
		}
		if ($sounds{$key}{list}) {
			&write_link_list('', '', @{ $sounds{$key}{list} });
		}
		&write_newline($fpwtext, 1);
	}

	1;
}

#
# int write_plugin_files(void)
# creates text entries for text-based plugin files if necessary.
#
sub write_plugin_files()
{
	my @files = keys(%plugins);
	my (%plugin, @lines, $file, $code, $head_pos, $text_pos);
	local $_;

	return 0 if (!@files);

	@files = sort @files;
	&log("Writing plugin entries...\n");

	foreach $file (@files) {
		%plugin = %{ $plugins{$file} };

		open(IN, "<$plugin{path}")
			|| die "Unable to open file: $plugin{path} ($!)";
		@lines = <IN>;
		close(IN);

		# a plugin file may be encoded in other coding system than sjis.
		# always convert it to sjis so that single-byte kana's are handled
		# properly.
		$_ = join('', @lines);
		$code = Jcode::getcode($_);
		if ('ascii' ne $code && 'sjis' ne $code) {
			$_ = Jcode::convert($_, 'sjis', $code);
		}
		@lines = split(/[\r\n]+/, $_);

		printf("$plugin{tag}: $file\n");
		$fpwtext->new_context()
			|| die &fpwerr("Unable to start new context", $fpwtext);
		($head_pos, $text_pos) = &new_entry($plugin{tag});

		$fpwheading->add_text("$file $LABEL_EUC{plugin}")
			|| die &fpwerr("Unable to add heading: $file $LABEL_EUC{plugin}",
					$fpwheading);
		$fpwkeyword->add_entry($file, $head_pos, $text_pos)
			|| die &fpwerr("Unable to add keyword: $file", $fpwkeyword);

		if (!$fpwtext->add_keyword_start()
				|| !$fpwtext->add_text($file)
				|| !$fpwtext->add_keyword_end()
				|| !$fpwtext->add_newline()) {
			die &fpwerr("Unable to write the first line", $fpwtext);
		}

		&write_link_list('', '', @{ $plugin{list} });
		&write_newline($fpwtext, 1);
		$fpwtext->add_indent_level(2);

		while (@lines) {
			$_ = shift(@lines);
			s/[\r\n\x1a]+$//;
			if ($_) {
				# a plugin file may be encoded in other coding system than
				# sjis. always convert it to sjis so that single-byte
				# kana's are handled properly.
				if ('ascii' ne $code && 'sjis' ne $code) {
					$_ = Jcode::convert($_, 'sjis');
				}
				&write_sjis($fpwtext, $_);
			}
			&write_newline($fpwtext, 1);
		}
	}

	scalar(@files);
}

#
# int write_plugin_ext(void)
# appends 'ext.csv' plugin if it exists in the source directory.
#
sub write_plugin_ext()
{
	return 0 if ('' eq $opt_plugin);

	my %ftype = {
		BINARY => 'oCi',
		TEXT => 'eLXg'
	};
	my %fcnt = {
		AUDIO => 'yEȂ',
		IMAGE => '摜',
		VIDEO => '',
		COMP => 'kf[^',
		ARC => 'k𔺂ȂA[JC',
		PROG => 'vO\[Xt@C',
		EXE => 'vOst@C',
		INF => '',
		DB => 'f[^x[X',
		TEXT => ''
	};
	my $file = &find_plugin('ext.csv') or return;
	my $handle = new IO::File($file, 'r')
		|| die "Unable to open file: $file ($!)\n";
	my $cnt = 0;
	my (@col, $text_pos, $head_pos, $tag);

	binmode $handle;
	&log("Plugin file 'ext.csv' found, start parsing...\n");
	$fpwtext->new_context()
		|| die &fpwerr("Unable to start new context", $fpwtext);

	while ($_ = $handle->getline()) {
		s/([^\r\n]*)[\r\n]+$/$1/;
		@col = split(/,/);
		next if ('' eq $col[0]);

		$cnt++;
		printf("EXT(%08x): $col[0]\n", $cnt);

		# add this line as a new entry.
		($head_pos, $text_pos) = &new_entry('EXT', $cnt);
		&write_sjis($fpwheading, "$col[0] ygqz");

		# register search words.
		if (!$fpwword2->add_entry($col[0], $head_pos, $text_pos)
				|| !$fpwword2->add_entry(".$col[0]", $head_pos, $text_pos)) {
			die &fpwerr("Unable to add search word: $col[0]", $fpwword2);
		}

		# write the first line.
		$fpwtext->add_keyword_start();
		&write_sjis($fpwtext, $col[0]);
		$fpwtext->add_keyword_end();
		&write_sjis($fpwtext, " ygqz\n");
		$fpwtext->add_indent_level(2);

		# file type.
		if ('' ne $col[1]) {
			if (defined($ftype{$col[1]})) { $_ = $ftype{$col[1]}; }
			else { $_ = $col[1]; }
		} else {
			$_ = '';
		}
		if ('' ne $col[2]) {
			$_ .= '/' if ('' ne $_);
			if (defined ($fcnt{$col[2]})) { $_ .= $fcnt{$col[2]}; }
			else { $_ .= $col[2]; }
		}
		if ('' ne $_) {
			$fpwtext->add_text($opt_bullet)
				|| die &fpwerr("Unable to add text: $opt_bullet", $fpwtext);
			&write_sjis($fpwtext, "$_\n");
		}

		# MIME type.
		if ('' ne $col[3]) { $_ = $col[3] }
		else { $_ = ''; }
		if ('' ne $col[4]) {
			$_ .= '/' if ('' ne $_);
			$_ .= $col[4];
		}
		if ('' ne $_) {
			$fpwtext->add_text($opt_bullet)
				|| die &fpwerr("Unable to add text: $opt_bullet", $fpwtext);
			&write_sjis($fpwtext, "MIME ^Cv: $_\n");
		}

		# description.
		if ('' ne $col[5]) {
			$fpwtext->add_text($opt_bullet)
				|| die &fpwerr("Unable to add text: $opt_bullet", $fpwtext);
			&write_sjis($fpwtext, &reform_sjis($col[5], 0) . "\n");
		}

		&write_newline($fpwtext, 1);
	}

	$handle->close();
	$cnt;
}

1;
