#
# wdic-reform.pl
# $Id: wdic-reform.pl 1.4.0.1 2004/07/06 07:52:52 satomii Exp $
#
# DO NOT CHANGE THE KANJI CODE OF THIS FILE.
#
# Copyright (C) 2000-2003, rei@wdic.org.
# This program is distributed in the hope that it will be useful, but
# without any warranty. See the GNU General Public License for the details.
#
use strict;
use vars qw(
	$SJIS_ANY
	$SJIS_X0201
);

$SJIS_ANY = '(?:[\x81-\x9f\xe0-\xfc][\x40-\x7e\x80-\xfc])';
$SJIS_X0201 = '[\xa1-\xdf]';

#
# regular expressions for re-formatting.
#
my $ASCII_EXCEPT_QUOTE = '[\x21\x23-\x7e]';

#  [łƂɂ͂] ('' ͕Kv?)
my $SJIS_PARTICLE = '(?:\x82[\xaa\xc5\xc6\xc9\xcd\xe0\xe2\xf0])';
my $SJIS_KATAKANA = '(?:\x83[\x40-\x96]|\x81\x5b)';
my $SJIS_GREEK = '(?:\x83[\x9f-\xd6])';

my $ASCII_NOSPACE_BOTH = '[\x00-\x20\"\'\-\+\*\/=_\\\\\@\|]';
my $ASCII_NOSPACE_BEFORE = '(?:[,\.!\?:;\)\]\>\}]|' . $ASCII_NOSPACE_BOTH . ')';
my $ASCII_NOSPACE_AFTER = '(?:[\(\[\<\{]|' . $ASCII_NOSPACE_BOTH . ')';

my $SJIS_NOSPACE_BOTH =
	# ABCDEFJK]^{|}~
	'(?:\x81[\x41-\x46\x4a\x4b\x5d\x5e\x7b-\x88\x8b-\x8d\x99'
	# ȁ́ځہ܁݁ށ߁
	. '\xb8-\xbf\xc8\xcd\xda-\xe7]|'
	# [-], [@--]
	. $SJIS_GREEK . '|\x84[\x40-\xbe])';
my $SJIS_NOSPACE_BEFORE =
	# GHILfhjln
	'(?:\x81[\x47-\x49\x4c\x66\x68\x6a\x6c\x6e'
	# prtvxz
	. '\x70\x72\x74\x76\x78\x7a\x8e\x93\x97\xf0-\xf3]|'
	. $SJIS_NOSPACE_BOTH . ')';
my $SJIS_NOSPACE_AFTER =
	# Megikmo
	'(?:\x81[\x46\x4d\x65\x67\x69\x6b\x6d\x6f'
	# qsuwy
	. '\x71\x73\x75\x77\x79\x8f-\x92\x97\x98\x9a-\x9e\x9f-\xa7]|'
	. $SJIS_NOSPACE_BOTH . ')';

my $NOSPACE_BOTH = "(?:$SJIS_NOSPACE_BOTH|$ASCII_NOSPACE_BOTH)";
my $NOSPACE_BEFORE = "(?:$SJIS_NOSPACE_BEFORE|$ASCII_NOSPACE_BEFORE)";
my $NOSPACE_AFTER = "(?:$SJIS_NOSPACE_AFTER|$ASCII_NOSPACE_AFTER)";

my $CHEM_SYMBOL = '(?:A[cglmrstu]|B[aehikr]?|C[adeflmorsu]?|D[by]'
	. '|E[rsu]|F[emr]?|G[ade]|H[efgos]?|I[nr]?|Kr?'
	. '|L[airu]|M[dgnot]|N[abdeiop]?|Os?|P[abdmortu]?|R[abefhnu]'
	. '|S[bcegimnr]?|T[abcehilm]|U(?:u[bhnopqstu])?|[VW]'
	. '|Xe|Yb?|Z[nr])';

my $MAYBE_COMP_PHRASE = '(?:[A-Z][a-z0-9\']+(?: [A-Za-z0-9][A-Za-z0-9\']*)+)';
my $MAYBE_ACRO = '(?:(?:[0-9]+th )?[A-Z][A-Za-z0-9\-]*[A-Za-z0-9])';
my $MAYBE_QUANT = '(?:[0-9]+(?:,[0-9]{3})*(?:.[0-9]+)? *'
	. '(?:[KGM]|[KGMT]?b(?:ps|ytes?)?|[mck]m(?:\^?[2-9][0-9]*)?|[mk]g)'
	. '(?:\/[a-z0-9\^]+)?\))';
my $MAYBE_IP = '(?:(?:[1-9][0-9]?[0-9]?|xx?x?)(?:\.(?:[1-9][0-9]?[0-9]?|xx?x?)){3})';

sub zen2han_sjis($$)
{
	my ($in, $mode) = @_;
	my @list = unpack('C*', $in);
	my $out = '';
	my $prev = '';
	my ($c1, $c2);

	while ($c1 = shift(@list)) {
		$c2 = 0;
		$in = substr($in, 1);
		if (0x81 == $c1) {
			$c2 = shift(@list);
			$in = substr($in, 1);
			if (0x40 == $c2) {							# '@'
				$c1 = 0x20; $c2 = 0;
			} elsif (0x48 == $c2) {						# H
				$c1 = 0x3f;
				if ($mode < 2 && '' ne $in && $in !~ /^$NOSPACE_BEFORE/) {
					$c2 = 0x20;
				} else {
					$c2 = 0;
				}
			} elsif (0x49 == $c2) {						# I
				$c1 = 0x21;
				if ($mode < 2 && '' ne $in && $in !~ /^$NOSPACE_BEFORE/) {
					$c2 = 0x20;
				} else {
					$c2 = 0;
				}
			} elsif (0x5e == $c2) {						# ^
				$c1 = 0x2f; $c2 = 0;
			} elsif (0x69 == $c2 || 0x6a == $c2) {		# ij
				$c1 = $c2 - 0x41; $c2 = 0;
			} elsif (0x66 == $c2) {						# f
				if (2 == $mode
						|| $in =~ /^[0-9][0-9][^0-9]*/
						|| $in =~ /^\x82[\x4f-\x58]\x82[\x4f-\x58]/) {
					$c1 = 0x27; $c2 = 0;
				}
			} elsif (0x93 == $c2) {						# 
				if (2 == $mode || $prev =~ /^[0-9]+$/) {
					$c1 = 0x25; $c2 = 0;
				}
			} elsif ($mode < 2) {
				if (0x46 == $c2) {						# F
					if ("\x97\xe1" eq $prev) {			# 
						$c1 = 0x3a; $c2 = 0x20;
					}
				} elsif (1 == $mode) {
					if (0x41 == $c2) {					# A
						$c1 = 0x2c; $c2 = 0x20;
					} elsif (0x42 == $c2) {				# B
						$c1 = 0x2e; $c2 = 0x20;
					}
				}
			} else {									# keyword mode only.
				if (0x43 == $c2) {						# C
					$c1 = 0x2c; $c2 = 0;
				} elsif (0x44 == $c2) {					# D
					$c1 = 0x2e; $c2 = 0;
				} elsif (0x46 == $c2 || 0x47 == $c2) {	# FG
					$c1 = $c2 - 0x0c; $c2 = 0;
				} elsif (0x62 == $c2) {					# b
					$c1 = 0x7c; $c2 = 0;
				} elsif (0x68 == $c2) {					# h
					$c1 = 0x22; $c2 = 0;
				} elsif (0x6d == $c2) {					# m
					$c1 = 0x5b; $c2 = 0;
				} elsif (0x6e == $c2) {					#  n
					$c1 = 0x5d; $c2 = 0;
				} elsif (0x7b == $c2) {					# {
					$c1 = 0x2b; $c2 = 0;
				} elsif (0x7c == $c2) {					# |
					$c1 = 0x2d; $c2 = 0;
				} elsif (0x96 == $c2) {					# 
					$c1 = 0x2a; $c2 = 0;
				} elsif (0x93 == $c2) {					# 
					$c1 = 0x25; $c2 = 0;
				} elsif (0x94 == $c2) {					# 
					$c1 = 0x23; $c2 = 0;
				}
			}
		} elsif (0x82 == $c1) {
			$c2 = shift(@list);
			$in = substr($in, 1);
			if ((0x4f <= $c2 && $c2 <= 0x58)
					|| (0x60 <= $c2 && $c2 <= 0x79)) {	# [O-X`-y]
				$c1 = $c2 - 0x1f; $c2 = 0;
			} elsif (0x81 <= $c2 && $c2 <= 0x9a) {		# [-]
				$c1 = $c2 - 0x20; $c2 = 0;
			}
		} elsif ((0x83 <= $c1 && $c1 <= 0x9f)
				|| (0xe0 <= $c1 && $c1 <= 0xfc)) {		# other SJIS.
			$c2 = shift(@list);
			$in = substr($in, 1);
		} elsif (0xa5 == $c1)	{						# E in single byte.
			if ($prev !~ /^[\xa1-\xdf]$/ && $in !~ /^[\xa1-\xdf]/) {
				$c1 = 0x81;
				$c2 = 0x45;
			}
		}
		$out .= pack('C', $c1);
		$out .= pack('C', $c2) if ($c2);
	}
	$out;
}

#
# char *reform_sjis(char *in, int mode)
# reformat SJIS string, inserting white spaces between double- and
# single-byte characters, converting some DBCS into SBCS. but this
# routine may still contain some improper operation...
#
sub reform_sjis($$)
{
	my ($in, $mode) = @_;
	my $out = '';
	my $prev = '';
	my (@quote, $str);

	$in = &zen2han_sjis($in, $mode);
	$in =~ s/  +/ /g;

	while ('' ne $in) {
		if ($in =~ s/^ +//) {
			next if ('' eq $prev || $prev =~ / $/ || $in =~ /^,/);
			$str = ' ';

		} elsif ($in =~ s/^((?:ZOg|zXg):$SJIS_ANY+)//) {
			# special case.
			$str = $1;
			if ('' ne $prev
					&& $prev !~ /$NOSPACE_AFTER$/
					&& $prev !~ /\<[a-z][^\<\>]*\>$/i
					&& $str !~ /^$SJIS_NOSPACE_BEFORE/) {
				$str = " $str";
			}

		} elsif ($in =~ s/^(~)([\-\+])($SJIS_ANY|[A-Za-z0-9])/$3/) {
			$str = "$1 $2 ";
			if ('' ne $prev && $prev !~ /$NOSPACE_AFTER$/) {
				$str = " $str";
			}

		} elsif ($in =~ s/^($SJIS_ANY+)//) {
			$str = $1;
			if ('' ne $prev
					&& $prev !~ /$NOSPACE_AFTER$/
					&& $prev !~ /\<[a-z][^\<\>]*\>$/i
					&& $str !~ /^$SJIS_NOSPACE_BEFORE/) {
				$str = " $str";
			}
			if (($str =~ /Ղ$/ || $str !~ /$SJIS_PARTICLE$/)
					&& $in =~ s/^(er)([^A-Za-z])/$2/) {
				$str = "$str$1";
			} elsif ('' ne $in
					&& $in !~ /^$ASCII_NOSPACE_BEFORE/
					&& $in !~ /^(\([nxy]\)|\<\/[a-z][^\<\>]*\>)/i
					&& $str !~ /$SJIS_NOSPACE_AFTER$/) {
				$str = "$str ";
			}

		} elsif ($in =~ s/^($SJIS_X0201+)//) {
			$str = $1;

		} elsif ($in =~ s/^(\'[0-9]{2})([^0-9])/$2/
				|| $in =~ s/(\'\x82[\x4f-\x58]\x82[\x4f-\x58])//) {
			# like '99 or 'XX, 2-digit year.
			$str = $1;
			if ($mode < 2
					&& '' ne $in
					&& '' ne $prev
					&& $prev !~ /$NOSPACE_AFTER$/) {
				$str = " $str";
			}

		} elsif ($in =~ s/^([\"\'])//) {
			$str = $1;
			if ('\'' eq $str
					&& $prev !~ /$SJIS_ANY$/
					&& ($prev =~ /[A-Z]$/i
					|| ($prev =~ /[0-9]$/ && $in =~ /^[\-0-9]/))) {
				# maybe within some western word, nop.
			} elsif (@quote && $str eq $quote[$#quote]) {
				# quotation closed.
				pop @quote;
				if ($mode < 2
						&& '' ne $in
						&& ($in =~ s/^ +//
						|| $in !~ /^$NOSPACE_BEFORE/)) {
					$str = "$str ";
				}
			} elsif ($prev =~ /[0-9]$/ && $in !~ /\"/) {
				# like 00" (unit), nop.
			} else {
				# quotation opened.
				push(@quote, $str);
				if ($mode < 2
						&& '' ne $prev
						&& $prev !~ /$NOSPACE_AFTER$/) {
					$str = " $str";
				}
			}

		} elsif ($in =~ s/^\)\(($SJIS_ANY)/$1/) {
			# special cases.
			if ('' ne $prev && $prev !~ /$NOSPACE_AFTER$/) {
				$str = ') (';
			} else {
				$str = ')(';
			}

		} elsif ($in =~ s/^(\([nxy]\)(?:$ASCII_EXCEPT_QUOTE|$SJIS_GREEK|\x81.)+)//i
				|| $in =~ s/^(\<[\/!\?\%]?[a-z][^\<\>]*\>)//i
				|| $in =~ s/^([\.,])([1-9][0-9]* *)/ $2/
				|| $in =~ s/^(!\?)([A-Z])/ $2/i) {
			# special cases.
			$str = $1;

		} elsif ($in =~ s/^(\([a-z]=[0-9],[0-9],[^\(\)]+)//
				|| $in =~ s/^(W[\-\+])(EB\x81\x5bN{\x83\x5d)/ $2/
				|| $in =~ s/^(::[0-9]+)($SJIS_ANY|[ \.,\(])/$2/) {
			# special cases.
			$str = $1;
			if ('' ne $prev && $prev !~ /[ \(\[\"\']$/) {
				$str = " $str";
			}

		} elsif ($in =~ s/^([1-9][0-9]?[0-9]?(?:,[0-9]{3})+$ASCII_EXCEPT_QUOTE*)//
				|| $in =~ s/^([a-z]*\[([a-z]: *)?[a-z][a-z0-9]\]$ASCII_EXCEPT_QUOTE*)//) {
			# a. 1,000,000
			# b. lim[t0]
			$str = $1;
			if ('' ne $prev
					&& $prev !~ /$NOSPACE_AFTER$/
					&& $str !~ /^$NOSPACE_BEFORE/) {
				$str = " $str";
			}
			if ('' ne $in
					&& $in !~ /^$NOSPACE_BEFORE/
					&& $str !~ /$NOSPACE_AFTER$/) {
				$str = "$str ";
			}

		} elsif ($in =~ s/^([a-z0-9]+)(\((?:$MAYBE_QUANT|$MAYBE_IP)\))//i
				|| $in =~ s/^([0-9]+[GM])(\([0-9]+(?:,[0-9]{3})*\)) *($SJIS_ANY)/ $3/
				|| $in =~ s/^($MAYBE_COMP_PHRASE)(\($MAYBE_ACRO\))($SJIS_ANY|[ \.,\"])/$3/
				|| $in =~ s/^(\($MAYBE_ACRO:)($MAYBE_COMP_PHRASE)//) {
			# a. (1000bps) | (xxx.xxx.xxx.xxx)
			# b. 999M (1,047,527,424) oCg
			# c. Full Spell (FS)
			# d. (FS: Full Spell)
			$str = "$1 $2";
			if ('' ne $prev && $prev !~ /$NOSPACE_AFTER$/) {
				$str = " $str";
			}
			if ('' ne $in && $in !~ /^$NOSPACE_BEFORE/) {
				$str = "$str ";
			}

		} elsif ($in =~ s/^(\([0-9][A-Za-z0-9]*\),?\-$ASCII_EXCEPT_QUOTE+)//
				|| $in =~ s/^([0-9]+(?:,[0-9]+)+\-)($SJIS_ANY)/$2/
				|| $in =~ s/^((?:[A-Za-z0-9]|BT)\-)($SJIS_KATAKANA+)/$2/) {
			# 1,2,3,4,5,6-wLT... etc.
			$str = $1;
			if ($prev =~ /$SJIS_ANY$/ && $prev !~ /$SJIS_NOSPACE_AFTER$/) {
				$str = " $str";
			}

		} elsif ($in =~ s/^([a-z0-9]+(?:, *[a-z0-9\']+)+,?) *($SJIS_ANY)/ $2/i) {
			$str = $1;
			$str =~ s/, */, /g;
			if ('' ne $prev && $prev !~ /$NOSPACE_AFTER$/) {
				$str = " $str";
			}

		} elsif ($in =~ s/^((?:[\-\+][0-9][0-9\.,]*|[0-9][0-9\.,]*[\-\+]))($SJIS_ANY)/$2/
				|| $in =~ s/^((?:C\/C\+\+|$CHEM_SYMBOL+[\-\+]+))(\(?$SJIS_ANY)/$2/
				|| $in =~ s/^((?:R\^|\*CD)\*)($SJIS_ANY)/$2/) {
			# a. -1 or 1- followed by SJIS char.
			# b. OH+ followed by SJIS char.
			# c. R^* or *CD* followed by SJIS char.
			$str = $1;
			if ($prev =~ /$SJIS_ANY$/ && $prev !~ /$SJIS_NOSPACE_AFTER$/) {
				$str = " $str";
			}
			if ($in !~ /$SJIS_NOSPACE_BEFORE/) {
				$str = "$str ";
			}

		} elsif ($in =~ s/^([a-z]+:[A-Za-z0-9_=\-\+\*\/\\\\\.,:;~!\?&#%\@]+)//
				|| $in =~ s/^((?:f(ile|tp)|https?|news):(?:$SJIS_ANY|[A-Za-z0-9_~:\/\.\-\+%@])+)//
				|| $in =~ s/^(\/[A-Za-z][A-Za-z0-9_~:\/\.\-\+\*%@\[\]]+)//) {
			# url, path or xml attribute.
			$str = $1;
			if ($prev =~ /$SJIS_ANY$/ && $prev !~ /$SJIS_NOSPACE_AFTER$/) {
				$str = " $str";
			}
			if ($in =~ /^($SJIS_ANY|\()/
					&& $in !~ /^$SJIS_NOSPACE_BEFORE/
					&& ($str =~ /\*$/ || $str !~ /$NOSPACE_AFTER$/)) {
				$str = "$str ";
			}

		} elsif ($in =~ s/^([A-Z0-9\-]+:[0-9]+)//i		# JIS X 0208:1983 etc.
				|| $in =~ s/^([A-Za-z0-9_=\-\+\^~]+(?:[\(\[](?:$ASCII_EXCEPT_QUOTE|$SJIS_GREEK|[\x81\x84].)+[\)\]]$ASCII_EXCEPT_QUOTE*)+)//
				|| $in =~ s/^([0-9]+\.[0-9]+\([0-9]+\))//
				|| $in =~ s/^([a-z0-9_\-\+\^~][a-z0-9_\.=\-\+\*\/%&\^~\\\\\@#\$]*)//i
				|| $in =~ s/^(\((?:$CHEM_SYMBOL[0-9,=\-\+\(\)]*)+\)$ASCII_EXCEPT_QUOTE*)//
				|| $in =~ s/^(\(!\(?$ASCII_EXCEPT_QUOTE*\)?\))//) {
			$str = $1;
			if ($prev =~ /$SJIS_ANY$/
					&& $prev !~ /$SJIS_NOSPACE_AFTER$/
					&& $str !~ /^$NOSPACE_BEFORE/) {
				$str = " $str";
			} elsif ($str =~ /^[\-\+]/ && $prev =~ /$/) {
				# spacial case in SCI.DIC.
				$str = " $str";
			}
			if ($in =~ /^$SJIS_ANY/
					&& $in !~ /^$SJIS_NOSPACE_BEFORE/
					&& $str !~ /$NOSPACE_AFTER$/) {
				$str = "$str ";
			}

		} elsif ($in =~ s/^(\()//) {
			$str = $1;
			if ('' ne $prev && $prev !~ /$NOSPACE_AFTER$/) {
				$str = " $str";
			}

		} elsif ($in =~ s/^([!\)])//) {
			$str = $1;
			if ('' ne $in && $in !~ /^$NOSPACE_BEFORE/) {
				$str .= ' ';
			}

		} elsif ($in =~ s/^\@//) {
			$str = '@';
			if (('' eq $prev || $prev =~ /\($/)
					&& $in =~ /^($SJIS_PARTICLE)/) {
				$str = "$str ";
			}
			if ($prev =~ /$SJIS_ANY$/
					&& $prev !~ /($SJIS_NOSPACE_AFTER|(O)?)$/) {
				$str = " $str";
			}

		} elsif ($in =~ s/^\|//) {
			$str = '|';
			if ($prev =~ /$SJIS_ANY$/
					&& $prev !~ /$SJIS_NOSPACE_AFTER$/
					&& $in =~ /^[\x21-\x7e]/) {
				$str= " $str";
			}

		} elsif ($in =~ s/^(\.$SJIS_ANY*)//) {
			$str = $1;

		} else {
			$str = substr($in, 0, 1);
			$in = substr($in, 1);
			if ($prev =~ /$SJIS_ANY$/
					&& $prev !~ /$SJIS_NOSPACE_AFTER$/
					&& $str !~ /^$ASCII_NOSPACE_BEFORE/) {
				$str = " $str";
			}
			if ($in =~ /^$SJIS_ANY/
					&& $in !~ /^$SJIS_NOSPACE_BEFORE/
					&& $str !~ /$ASCII_NOSPACE_AFTER$/) {
				$str = "$str ";
			}
		}

		$out .= $str;
		$prev = $str;
	}

	# (R), registered trade marks.
	$out =~ s/(Ce|vpeB) (\(R\))/$1$2/;
	# gq(||||||)?.foo
	$out =~ s/(gq$SJIS_PARTICLE?)(\.[A-Za-z0-9]+)/$1 $2/g;
	# "???.to" in GEO.DIC.
	$out =~ s/($SJIS_ANY)(\?\?\?+\.[a-z]+)/$1 $2/g;
	# in SERV.DIC etc.
	$out =~ s/(Richard M\.)(Stallman)/$1 $2/g;
	# in SUBS.DIC etc.
	$out =~ s/( (?:\(?$CHEM_SYMBOL[0-9=\-\+]*\)?[0-9]*)+[A-Za-z]*[\-\+])([\x82\x88-\x9f\xe0-\xfc].|CI)/$1\n$2/g;
	$out =~ s/\-\n($SJIS_NOSPACE_BEFORE)/\-$1/g;
	$out =~ s/\n/ /g;
	# in TELM.DIC etc.
	$out =~ s/( [A-Z][A-Z\-]*|$SJIS_ANY\-[ab])(\([A-Za-z][A-Za-z0-9\-]*\),? )/$1 $2/g;
	# in KNOW.DIC.
	$out =~ s/(\x83\x56\x83\x74\x83\x67 JIS)(\(Shift_JIS\))/$1 $2/ig;
	$out =~ s/( MessagePad)(\(Newton\) )/$1 $2/ig;
	# in LAPL.DIC.
	$out =~ s/($SJIS_ANY)(\.[Zz] )/$1 $2/;
	$out =~ s/(\<%=)(` %\>)/$1 $2/;
	# in SCI.DIC.
	$out =~ s/( \((?:$CHEM_SYMBOL[0-9\-\+]*)+\))([0-9]+mol)/$1 $2/;
	# in TELT.DIC.
	$out =~ s/(NT2:)(PBX\/LAN)/$1 $2/;
	$out =~ s/\*y/ * y/;
	# in USTAR.DIC.
	$out =~ s/;([1-9][0-9]+ N)/; $1/g;

	$out;
}

1;
